#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

void JacobiSolve(nx, ny, sk, id, ik, f, u )
int    nx,ny,*id,*ik;
double *sk, *f, *u;
{
  extern double dscapr();
  double *dd, *r, dzero=0.0, sigma, sigma0, tol, tol2, omega;
  int nnode, iter, maxiter;
  
  omega   = 1.0;
  maxiter = 100;
  tol  = 1e-5;						/* tolerance   */
  tol2 = tol*tol;					/* tolerance^2 */
  
  nnode = (nx+1)*(ny+1); 
  dd    = (double *) malloc(nnode*sizeof(double));  /* matrix diagonal */
  r     = (double *) malloc(nnode*sizeof(double));  /* residual */

  GetDiag(nx, ny, sk, id, dd); 		/* Get diag */
	 
  /*  Choose initial guess  */
  vdcopy(nnode, u, 1, &dzero, 0);		/*  u := 0         */

  /*  Initial sweep  */
  vdcopy(nnode, r, 1, f, 1);			/*                 */
  CrsMult(0, nnode-1, r, u, id, ik, sk, -1.0);	/*  r := f - K*u   */
  
  vddiv(nnode, r, 1, r, 1, dd, 1);		/*  r := D^{-1}*r  */
  sigma0 = dscapr(nnode,r,r);			/* s0 := <r,r>     */   
  
    printf("sigma : %g\n",sigma0);     
  /* Iteration sweeps */
  iter  = 0;
  sigma = sigma0;
  while ( sigma > tol2*sigma0 && maxiter>iter)
  {
    iter++;
    vdaxpy(nnode, u, u, omega, r );		/*  u := u + om*r  */  
    vdcopy(nnode, r, 1, f, 1);			/*                 */
    CrsMult(0, nnode-1, r, u, id, ik, sk, -1.0);/*  r := f - K*u   */

    vddiv(nnode, r, 1, r, 1, dd, 1);		/*  r := D^{-1}*r  */ 
    sigma = dscapr(nnode,r,r);			/* s0 := <r,r>     */
    
    printf("Iteration %u : %g\n",iter,sqrt(sigma/sigma0));     
  }
  	printf("aver. Jacobi rate :  %g\n", exp(log(sqrt(sigma/sigma0))/iter));
 
  free(r); free(dd);
  return;
}
