#include <stdio.h>
#include <math.h>
#include "mg.h"

void MGM(pp, crtl, level, neigh)
PointerStruct 	*pp;
ControlStruct 	*crtl;
int       level, neigh[4];
{
 double *sk, *dd, *u, *f, *d, *w, *aux, *d_c, *w_c, dzero=0.0;
 int    *id, *ik, nx, ny, nnode, nx_c, ny_c, j;

 nx = pp->nx[level];
 ny = pp->ny[level];
 sk = pp->sk[level];
 id = pp->id[level];
 ik = pp->ik[level];
 dd = pp->dd[level];
 u  = pp->u[level];
 f  = pp->f[level];
 aux= pp->aux[level];			/* all fixed fine matrices/arrays */
 nnode = (nx+1)*(ny+1);
 d  = aux;				/* defect on fine grid		*/
 w  = aux;				/* correction on fine grid	*/

 if (level==0) 
 { 
  			/* printf("-------------\nCoarse grid solver\n"); */
  JacobiSolve2(nx, ny, sk, id, ik, f, u, dd, aux );
  			/* printf("-------------\n");			*/
 }
 else
 {

  nx_c = pp->nx[level-1];
  ny_c = pp->ny[level-1]; 
  d_c  = pp->f[level-1];		/* defect on corse grid		*/
  w_c  = pp->u[level-1];		/* correction on coarse grid	*/

 					/* Presmoothing			*/
  			/* printf("%u : Smooth-Pre\n",level);		*/
  JacobiSmooth(nx, ny, sk, id, ik, f, u, dd, aux, crtl->npre );
 
  			/* printf("%u : Mult\n",level);			*/
  vdcopy(nnode, d, 1, f, 1);		/* d := f - K*u   		*/
  CrsMult(0, nnode-1, d, u, id, ik, sk, -1.0);

  			/* printf("%u : Restrict\n",level);		*/
  Restrict(nx, ny, d,  nx_c, ny_c, d_c, neigh);
 
  vdcopy((nx_c+1)*(ny_c+1), w_c, 1, &dzero, 0);		/*  w_c := 0	*/

  			/* printf("%u : MGM-cycle\n",level);		*/
  for (j=1; j<=crtl->cycle; j++)
       MGM(pp, crtl, level-1, neigh);
  
  			/* printf("%u : Interpolate\n",level);		*/
  Interpolate(nx, ny, w,  nx_c, ny_c, w_c, neigh);

  vdaxpy(nnode, u, u, 1.0, w );		/*  u := u + w			*/  

  
					/* Postsmoothing		*/
  			/* printf("%u : Smooth-Post\n",level);		*/
  JacobiSmooth(nx, ny, sk, id, ik, f, u, dd, aux, crtl->npost );
 
 }

 return;
}


void MGMSolver(pp, crtl, level, neigh)
PointerStruct 	*pp;
ControlStruct 	*crtl;
int       level, neigh[4];
{
 extern double dscapr();
 double *sk, *u, *f, *aux, *r, dzero=0.0, sigma0, sigma, tol, tol2; 
 int    *id, *ik, nx, ny, nnode, iter, maxiter, *fl;

 				printf("In MGMSolver\n");
 
 maxiter = 20;
 tol  = 1e-4;						/* tolerance   */
 tol2 = tol*tol;					/* tolerance^2 */

 nx = pp->nx[level];
 ny = pp->ny[level];
 sk = pp->sk[level];
 id = pp->id[level];
 ik = pp->ik[level];
 u  = pp->u[level];
 f  = pp->f[level];
 aux= pp->aux[level];			/* all fixed fine matrices/arrays */
 nnode = (nx+1)*(ny+1);
 r  = aux;				
 
/* vdcopy(nnode,u,1,&dzero,0);	*/		/*  u := 0	   */
 
 vdcopy(nnode, r, 1, f, 1);			/*                 */
 CrsMult(0, nnode-1, r, u, id, ik, sk, -1.0);	/*  r := f - K*u   */
 
 sigma0 = dscapr(nnode,r,r);			/* s0 := <r,r>     */   
 
  /* Iteration sweeps */
 iter  = 0;
 sigma = sigma0;
 while ( sigma > tol2*sigma0 && maxiter>iter)
 {
   iter++;
   
   MGM(pp, crtl, level, neigh); 

   vdcopy(nnode, r, 1, f, 1);			/*                 */
   CrsMult(0, nnode-1, r, u, id, ik, sk, -1.0);	/*  r := f - K*u   */

   sigma = dscapr(nnode,r,r);			/* s0 := <r,r>     */   

   	printf("MG-Iteration  %u : %g\n",iter,sqrt(sigma/sigma0));     
 }
 
 	printf("aver. MG rate :  %g\n", exp(log(sqrt(sigma/sigma0))/iter));

 return;
}
