C***************************************************************************
C*
C*	SUBROUTINE Send_ProcD(to,nin,xin)
C*
C*	--->	to		: I 
C*	--->	nin		: I
C*	<---	xin(nin)	: D
C*
C*	Sends "nin" double precision numbers stored in "xin" to process "to".
C*	Runs also on heterogeneous networks.
C*
C***************************************************************************

	SUBROUTINE Send_ProcD(to,nin,xin)

	IMPLICIT NONE

        INCLUDE 'mpif.h'
	DOUBLE PRECISION xin
	INTEGER	  to,nin,tag,ierr
	DIMENSION xin(nin)
	
C					Mark the message (other choices !)
	tag = 0
C					Send length of the data

	CALL MPI_Send(nin,1,MPI_INTEGER,to,tag,MPI_COMM_WORLD,ierr)
				IF (ierr.NE.0) GOTO 99
	IF (nin.NE.0) THEN
C					Send the data
     	
	  CALL MPI_Send(xin,nin,MPI_DOUBLE_PRECISION,to,tag,
     &					MPI_COMM_WORLD,ierr)
				IF (ierr.NE.0) GOTO 99
	ENDIF
	
	RETURN
   99	WRITE(*,*) 'Send_ProcD : Error ',ierr
   	CALL MPI_FINALIZE(ierr)
	END

C***************************************************************************
C*
C*	SUBROUTINE Recv_ProcD(from,nout,xout,maxbuf)
C*
C*	--->	from		: I
C*	<---	nout		: I
C*	<---	xout(nout)	: D
C*	--->	maxbuf		: I 		nout <= maxbuf !!
C*
C*	Recveives "nout" double precision numbers stored  in "xout" 
C*	from process "from".
C*	Runs also on heterogeneous networks.
C*
C***************************************************************************

	SUBROUTINE Recv_ProcD(from,nout,xout,maxbuf)

	IMPLICIT NONE

        INCLUDE 'mpif.h'
	DOUBLE PRECISION xout
	INTEGER	  from,nout,maxbuf,tag,ierr,status
	DIMENSION xout(maxbuf),status(MPI_STATUS_SIZE)
	
C				Expected tag of the message (other choices !)
	tag = MPI_ANY_TAG
C					Receive length of the data

	CALL MPI_Recv(nout,1,MPI_INTEGER,from,tag,
     &					MPI_COMM_WORLD,status,ierr)
				IF (ierr.NE.0) GOTO 99
	IF (nout.NE.0) THEN
C					Buffer xout too small
	  IF (nout.GT.maxbuf) THEN
	  	WRITE(*,*) 'Buffer too small in Recv_ProcD'
	  	GOTO 99
	  ENDIF
C					Receive the data
     	
	  CALL MPI_Recv(xout,nout,MPI_DOUBLE_PRECISION,
     &			from,tag,MPI_COMM_WORLD,status,ierr)
				IF (ierr.NE.0) GOTO 99
	ENDIF
	
	RETURN
   99	WRITE(*,*) 'Recv_ProcD : Error ',ierr, '  Status ',status
	CALL MPI_FINALIZE(ierr)
	END

C***************************************************************************
C*
C*	SUBROUTINE ExchangeD(yourid,nin,xin,nout,xout,maxbuf)
C*
C*	--->	yourid		: I 
C*	--->	nin		: I
C*	<---	xin(nin)	: D
C*	--->	from		: I
C*	<---	nout		: I
C*	<---	xout(nout)	: D
C*	--->	maxbuf		: I 		nout <= maxbuf !!
C*
C*	Exchanges double precision data with process yourid. 
C*	The data to send in "xin" and the received one in "xout" have 
C*	to be stored in different arrays !
C*	Runs also on heterogeneous networks.
C*
***************************************************************************

	SUBROUTINE ExchangeD(yourid,nin,xin,nout,xout,maxbuf)

	IMPLICIT NONE

        INCLUDE 'mpif.h'
	DOUBLE PRECISION xin,xout
	INTEGER	  yourid,nin,nout,maxbuf,myid,ierr,i
	DIMENSION xin(nin),xout(maxbuf)

C			yourid not admissible ?
	IF (yourid.LT.0) THEN
	   nout = -1
	   GOTO 999
	ENDIF
C			Get my rank "myid" in the group

	CALL MPI_COMM_RANK( MPI_COMM_WORLD, myid, ierr )
				IF (ierr.NE.0) GOTO 99

C			Sender and receiver the same ?
	IF (yourid.EQ.myid) THEN
	   nout = nin
	   DO i=1,nout
	     xout(i) = xin(i)
	   END DO
           GOTO 999
        ENDIF 
C			We need some unique choice which of the two
C			processes {myid|yourid} first sends or receives.

	IF (myid.LT.yourid) THEN
		CALL Send_ProcD(yourid,nin,xin)
		CALL Recv_ProcD(yourid,nout,xout,maxbuf)
	  ELSE
		CALL Recv_ProcD(yourid,nout,xout,maxbuf)
		CALL Send_ProcD(yourid,nin,xin)	
	ENDIF	

  999	RETURN
   99	CALL MPI_FINALIZE(ierr)
   	GOTO 999
	END
