/* ********************************************************************	*/
/*				M A S T E R 				*/
/* ********************************************************************	*/
#include <stdio.h>
#include "mpi.h"

/*----------------------------------------------------------------------*/
#define DEAD 1
#define ACTIVE 0

typedef struct 
{
  int start_index, end_index;	/*  Start and end of data block		*/
  int flag;			/*  Process ACTIVE or DEAD ?		*/
  int count;			/*  Number of handled data blocks 	*/
}
protocol;

/*----------------------------------------------------------------------*/

void calc_master(myid,num,icomm)
MPI_Comm icomm;			/*  Communicator			*/
int      num,			/*  # processes in icomm		*/
         myid;			/*  Own	rank in icomm (=0 for master)	*/
{
 int	    datalength, 		/*  Lenght of huge data array	*/
 	    maxlength,			/*  Max. length of data to send	*/ 
            stride,			/*  Stride for huge data array	*/
	    nslaves,			/*  # slaves processes 		*/
	    nbusy,			/*  # of busy/active slaves	*/
 	    i,ip,tag,info,source;
 double	    *data,			/*  pointer to data array	*/
            s;
 protocol   *prot;			/*  pointer to slave protocols	*/
 MPI_Status stat;

 nslaves    = num-1;
 datalength = 1000*nslaves;
 data       = (double*)malloc(datalength*sizeof(double));
 for (i=0;i<datalength;i++) data[i]=i+1;

/*		Broadcast maximal length of messages			*/
 maxlength  = datalength/100/nslaves;
 MPI_Bcast(&maxlength,1,MPI_INT,0,icomm);

/* 		Allocate protocal array, (prot[0] is unused)		*/
 prot       = (protocol*)malloc((nslaves+1)*sizeof(protocol));
 for (i=0;i<=nslaves;i++)
   { prot[i].start_index = 0; 
     prot[i].end_index   = -1;
     prot[i].flag        = DEAD;
     prot[i].count       = 0;
   }

/*		Initialize slave data					*/ 
 
 stride = maxlength;
 ip   = 0;
 nbusy = 0;
 tag  = 1111;
 for (i=1; i<=nslaves; i++)
  {
    info = MPI_Send(data+ip,stride,MPI_DOUBLE,i,tag,icomm);
    if (info==MPI_SUCCESS)
       { prot[i].start_index = ip; 
         prot[i].end_index   = ip+stride-1; 
         prot[i].flag        = ACTIVE;
	 prot[i].count++;
	 nbusy++;
        }
    ip  += stride;
  }

 while (nbusy>0)
 {
/*		Receive result from any slave				*/
    tag = 2222;
    MPI_Recv(&s,1,MPI_DOUBLE,MPI_ANY_SOURCE,tag,icomm, &stat);
    source = stat.MPI_SOURCE;
/*    printf("%d : %f\n",source,s); */

    if (ip<datalength)
     {
/*		Send that slave new data				*/    
       tag = 3333;
       info = MPI_Send(data+ip,stride,MPI_DOUBLE,source,tag,icomm);
       if (info==MPI_SUCCESS)
	  { prot[source].start_index = ip; 
            prot[source].end_index   = ip+stride-1; 
            prot[source].flag        = ACTIVE;
	    prot[source].count++;
	  }    
       ip += stride;
     }
    else  
     {
/*		Send that slave the end-of-computation flag		*/    
       tag = 3333;
       MPI_Send(&s,0,MPI_DOUBLE,source,tag,icomm);
       prot[source].flag = DEAD;
       nbusy--;
     }
 }

printf(">> %d  %d %d %d\n",prot[1].end_index,prot[2].end_index,
			   prot[3].end_index,prot[4].end_index);
printf(">> %d  %d %d %d\n",prot[1].count,prot[2].count,
			   prot[3].count,prot[4].count);
 
  MPI_Barrier(icomm);
  fflush(stdout); 
}
